import fizz
import asyncio
import threading
import queue

# define a queue that can be awaited on 
class AsyncQueue:

    def __init__(self,loop):
        self.event = asyncio.Event(loop = loop)
        self.queue = queue.SimpleQueue()

    def push(self,item):
        self.queue.put(item)
        self.event.set()
    
    def pull(self):
        if self.queue.empty():
            return None
        else:
            item = self.queue.get()
            if self.queue.empty():
                self.event.clear()
            return item

    async def wait(self):
        await self.event.wait()

# define an Actor that provides async methods
class AsyncActor(fizz.Actor):
    
    def __init__(self,label):
        self.label = label
        self.tasks = {}

    # async version of the query method
    async def query(self,options,gist,weight,mood):
        uid = super().query(options,gist,weight,mood)
        queue = AsyncQueue(self.loop)
        self.tasks[uid] = [queue,0]
        while True:
            await queue.wait()
            chunk = queue.pull()
            yield chunk
            if not(chunk):
                break 

    def onInit(self):
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self.onLoop, args=(), daemon=True)
        self.thread.start()
        asyncio.run_coroutine_threadsafe(self.onInitCO(),self.loop)

    def onExit(self):
        self.loop.stop()

    def onLoop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()   
        self.loop.close()  

    def onBeat(self):
        old = []
        for uid in self.tasks.keys():
            self.tasks[uid][1] = self.tasks[uid][1] + 1
            if self.tasks[uid][1] > 3:
                old.append(uid)
        for uid in old:
            self.tasks[uid][0].push(None)
            self.tasks.pop(uid)
        asyncio.run_coroutine_threadsafe(self.onBeatCO(),self.loop)
      
    def onTick(self,tick,chunks,ancillary):
        asyncio.run_coroutine_threadsafe(self.onTickCO(tick,chunks,ancillary),self.loop)

    def onReply(self,uid,chunk):
        if uid in self.tasks:
            self.tasks[uid][0].push(chunk)
        asyncio.run_coroutine_threadsafe(self.onReplyCO(uid,chunk),self.loop)

    def onKudos(self,uid,chunk):
        asyncio.run_coroutine_threadsafe(self.onKudosCO(uid,chunk),self.loop)

    # methods that can be implemented by derived class 

    async def onInitCO(self):
        pass

    async def onBeatCO(self):
        pass

    async def onTickCO(self,tick,chunks,ancillary):
        pass

    async def onReplyCO(self,uid,chunk):
        pass

    async def onKudosCO(self,uid,chunk):
        pass

# Implement an asynchronous processor which chains two queries in reaction to a keypress 'tweet' to compute a value
class Prober(AsyncActor):

    async def onTickCO(self,tick,chunks,ancillary):
        for chunk in chunks:
            if chunk[3] == 'tweet' and isinstance(chunk[4],fizz.Func):
                if chunk[4].label == 'keypress' and chunk[4].terms[0] == 100:
                    async for reply in self.query({'retries':4},fizz.Func('add',[3,2]),1,1):
                        if reply:
                            self.kudos(reply[2],{'retries':4},1,1,1) # we send a kudos back to whoever sent us the reply
                            v = reply[4]
                            async for reply in self.query({'retries':4},fizz.Func('add',[8,2]),1,1):
                                if reply:
                                    self.kudos(reply[2],{'retries':4},1,1,1) # we send a kudos back to whoever sent us the reply
                                    w = reply[4]
                                    print("result = " + str(v * w))
                                    self.tweet({'retries':4},v*w,1,1)
                                else:
                                    break
                        else:
                            #print("Timed out")
                            break
